/*
        file: nntp.c
          by: Johan Svensson <johan@ecsdg.lu.se>
    email to: <NA@ecsdg.lu.se>
        date: 1994-08-02
 last update: 1995-01-29
   copyright: Copyright (C) 1995-94 JoS-Ware Comp Tech, Sweden
     version: n/a
       notes: NNTP routines for NewsArchiver program

#      This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program; if not, write to the Free Software
#      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#


*/

#include "generic.h"
#include "nntp.h"


static FILE *ser_rd_fp = NULL;
static FILE *ser_wr_fp = NULL;


/* NNTP functions: */


SWORD socket_connect(char *machine) {

   SDWORD s;
   struct sockaddr_in sin;

   struct servent *sp;
   struct hostent *hp;

   if ((sp = getservbyname("nntp", "tcp")) ==  NULL) {
#ifdef VERBOSE
      (void) fprintf(stderr, "nntp/tcp: Unknown service.\n");
#endif
      return (-1);
   }

   if ((hp = gethostbyname(machine)) == NULL) {
#ifdef VERBOSE
      (void) fprintf(stderr, "%s: Unknown host.\n", machine);
#endif
      return (-1);
   }

   (void) memset((char *) &sin, 0, sizeof(sin));
   sin.sin_family = hp->h_addrtype;
   sin.sin_port = sp->s_port;

   /*
    * The following is kinda gross.  The name server under 4.3
    * returns a list of addresses, each of which should be tried
    * in turn if the previous one fails.  However, 4.2 hostent
    * structure doesn't have this list of addresses.
    * Under 4.3, h_addr is a #define to h_addr_list[0].
    * We use this to figure out whether to include the NS specific
    * code...
    */

   if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) { /* Get the socket */
#ifdef VERBOSE
      (void) perror("socket");
#endif
      return (-1);
   }

   /* And then connect */

   (void) memcpy((char *) &sin.sin_addr, hp->h_addr, hp->h_length);
   if (connect(s, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
#ifdef VERBOSE
      (void) perror("connect");
#endif
      (void) close(s);
      return (-1);
   }

   return (s);
}


SWORD put_server(char *string) {

   (void) fprintf(ser_wr_fp, "%s\r\n", string);
   (void) fflush(ser_wr_fp);
   return (0);
}

SWORD get_server(char *string, SDWORD size) {

   register char *cp;

   #ifdef DEBUGNNTP
   printf("get_server: start\n");
   #endif

   if (fgets(string, size, ser_rd_fp) == NULL)
      return (1);

   if ((cp = (char *)strchr(string, '\r')) != NULL)
      *cp = '\0';
   else if ((cp = (char *)strchr(string, '\n')) != NULL)
      *cp = '\0';

   #ifdef DEBUGNNTP
   printf("get_server: end\n");
   #endif

   return (0);
}


void close_server(void) {

   char    ser_line[256];

   #ifdef DEBUGNNTP
   printf("close_server: start\n");
   #endif

   if (ser_wr_fp == NULL || ser_rd_fp == NULL) {

      #ifdef DEBUGNNTP
      printf("close_server: file not open\n");
      #endif

      return;
   }

   put_server("QUIT");
   (void) get_server(ser_line, sizeof(ser_line));

   (void) fclose(ser_wr_fp);
   (void) fclose(ser_rd_fp);

   #ifdef DEBUGNNTP
   printf("close_server: end\n");
   #endif

   return;
}



SWORD connect_server(char *machine) {

   SWORD response;
   UWORD count = 2, max = 2;

   #ifdef DEBUGNNTP
   printf("connect_server: start\n");
   #endif

   response = nntp_connect(machine);
   if (response < 0) {
      #ifdef VERBOSE
         printf("Failed to connect to news server %s, retrying...\n", 
            machine);
      #endif
      while (count) {
         response = nntp_connect(machine);
         if (response < 0)  
            sleep(((max - count--) + 1) * 3);
         else
            break;
      }
      response = 0;
      if (!count) {
         #ifdef VERBOSE
            printf("Failed to connect to news server %s, aborting!\n", 
               machine);
         #endif
         response = -1;
      }
   }

   #ifdef DEBUGNNTP
   printf("connect_server: end\n");
   #endif

   return (response);
}



SWORD reconnect_server(char *machine) {

   SWORD retvalue;

   #ifdef DEBUGNNTP
   printf("reconnect_server: start\n");
   #endif

   close_server();
   retvalue = connect_server(machine);

   #ifdef DEBUGNNTP
   printf("reconnect_server: end\n");
   #endif

   return retvalue;
}


SWORD nntp_connect(char *machine) {

   SDWORD sockt_rd, sockt_wr;
   SDWORD status, alive = 1;
   char  *c, version[16];
   char  line[256], isnnrp[256];
   char  tempstr[255];


   #ifdef DEBUGNNTP
   printf("nntp_connect: start\n");
   #endif

   sockt_rd = socket_connect(machine);

   if (sockt_rd < 0)
      return (-1);

    /*
     * Now we'll make file pointers (i.e., buffered I/O) out of
     * the socket file descriptor.  Note that we can't just
     * open a fp for reading and writing -- we have to open
     * up two separate fp's, one for reading, one for writing.
     */

   #ifdef DEBUGNNTP
   printf("nntp_connect: open sockt_rd\n");
   #endif

   if ((ser_rd_fp = fdopen(sockt_rd, "r")) == NULL) {
      #ifdef VERBOSE
         perror("server_init: fdopen #1");
      #endif
      return (-1);
   }

   #ifdef DEBUGNNTP
   printf("nntp_connect: open sockt_wr\n");
   #endif

   sockt_wr = dup(sockt_rd);
   if ((ser_wr_fp = fdopen(sockt_wr, "w")) == NULL) {
      #ifdef VERBOSE
         perror("server_init: fdopen #2");
      #endif
      ser_rd_fp = NULL;               /* from above */
      return (-1);
   }

   #ifdef DEBUGNNTP
   printf("nntp_connect: setsockopt\n");
   #endif

   /* Set the socket status to avoid stale socket connections. */ 
   if (setsockopt(sockt_wr, SOL_SOCKET, SO_KEEPALIVE, 
            (char *)&alive, sizeof(alive)) < 0) {
      #ifdef VERBOSE
         printf("Failed to set KEEPALIVE for socket write!\n");
      #endif
   }
   alive = 1;
   if (setsockopt(sockt_rd, SOL_SOCKET, SO_KEEPALIVE,
            (char *)&alive, sizeof(alive)) < 0) {
      #ifdef VERBOSE
         printf("Failed to set KEEPALIVE for socket read!\n");
      #endif
   }

   /* Now get the server's signon message and check the revision on
    * of the server. Old servers don't have the distribution features
    * that WWW NNTP uses, so that feature off. 
    */

   strnset(line, '\0', strlen(line));

   #ifdef DEBUGNNTP
   printf("nntp_connect: get_server\n");
   #endif


   (void) get_server(line, sizeof(line));

   #ifdef DEBUGNNTP
   printf("nntp_connect: sscanf(line)\n");
   #endif

   sscanf(line, "%d", &status);

   #ifdef DEBUGNNTP
   printf("nntp_connect: check status\n");
   #endif

   if (status != 200 && status != 201) {
      if (strlen(line)) {
         #ifdef VERBOSE
            printf("Failed message from server is:\n %s\n", line);
         #endif
      }
      return(-1);
   }

   if (status == 201) {
/*
      Hey, don't print out this:
      printf("Server %s does not allow posting!\n", machine);
*/

      #ifdef DEBUGNNTP
      printf("nntp_connect: strstr(line, \"NNRP\"), status = %d\n", status);
      printf("nntp_connect: line = %s\n", line);
      #endif

      (void) sprintf(isnnrp, "NNRP");
      if (strstr(line, isnnrp) != NULL) {

         #ifdef DEBUGNNTP
         printf("nntp_connect: put_server\n");
         #endif

         put_server("XPAT");
         (void)get_server(line, sizeof(line));

         /* Just in case we're on a local INN server: */
         #ifdef DEBUGNNTP
         printf("nntp_connect: setting MODE READER\n");
         #endif

         sprintf(tempstr, "MODE READER");
         put_server(tempstr);
         /* Don't care about the reply */
         (void)get_server(line, sizeof(line));

         return (atoi(line));
      }
   }

   #ifdef DEBUGNNTP
   printf("nntp_connect: obviously 200 returned\n");
   #endif

   /* Just in case we're on a local INN server: */
   #ifdef DEBUGNNTP
   printf("nntp_connect: setting MODE READER\n");
   #endif

   sprintf(tempstr, "MODE READER");
   put_server(tempstr);
   /* Don't care about the reply */
   (void)get_server(line, sizeof(line));


   #ifdef DEBUGNNTP
   printf("nntp_connect: strstr before\n");
   #endif

   c = (char *)strstr(line, "version");

   #ifdef DEBUGNNTP
   printf("nntp_connect: strstr after\n");
   #endif

   if (c == NULL)
      return (atoi(line));
   sscanf(c, "%*s%s", version);

   #ifdef DEBUGNNTP
   printf("nntp_connect: hi again\n");
   #endif


   if (atoi(&version[2]) > 5 || atoi(&version[0]) > 1) {
      return (atoi(line));
   }
   if ((int)strlen(version) > 4 && atoi(&version[4]) > 3) {
   }
   else {
      #ifdef VERBOSE
         printf("NNTP version is %s, you need > \
                 1.5.3 for extended LIST commands!\n", version);
         printf("Group descriptions feature and distribution \
                 menu turned off!\n");
      #endif
   }

   #ifdef DEBUGNNTP
   printf("nntp_connect: end\n");
   #endif

   return (atoi(line));
}



void getauthorlist(char *machine, char *subjects[], 
     SDWORD first, SDWORD last, SDWORD row) {

   static char command[MAX_COMMAND_LEN], message[MAX_MESSAGE_LEN];
   char *author, *end, *brackbeg, *brackend, *strchr();
   SDWORD count = 0, status;

    
   (void) sprintf(command, "XHDR from %d-%d", first, last);
   put_server(command);
   get_server(message, sizeof(message));
   sscanf(message, "%i", &status);
   if (status != OK_HEAD) {
      reconnect_server(machine);
      put_server(command);
      get_server(message, sizeof(message));
      sscanf(message, "%i", &status);
      if (status != OK_HEAD)
         return;
   }
		

   for(;;) {

      get_server(message, sizeof(message));
	
      if (*message == '.') {
         break;
      }

      /*
       * message is of the form:
       *
       *    Number Author
       *
       *    201 johan@ecsdg.lu.se (Johan Svensson)
       *    202 Johan Svensson <johan@ecsdg.lu.se>
       *    203 <johan@ecsdg>
       *    302 "Johan Svensson" <johan@ecsdg.lu.se>
       *
       * must get the number since not all authors will be returned
       */

      /* Can be made fancyer at the expence of extra cpu time */
      author = strchr(message, ' ');
      assert(author != NIL(char));
      author++;

      /* First check for case 1, user@domain ("name") -> name */

      brackbeg = strchr(message, '(');
      brackend = strchr(message, '\0') - sizeof(char);
      /* brackend now points at the last ')' if this is case 1 */
      if (brackbeg != NIL(char) && (brackend > brackbeg) &&
            (*brackend == ')')) {
         author = brackbeg + sizeof(char);

	 /* Remove surrounding quotes ? */
	 if ((*author == '"') && (*(brackend - sizeof(char)) == '"')) {
	    author++;
	    brackend--;
	 }

	 /* Rather strip trailing spaces here */

	 *brackend = '\0';
      }
      else {
         /* Check for case 2, "name" <user@domain> -> name */
	 brackbeg = strchr(message, '<');
	 if (brackbeg != NIL(char) && (strchr(brackbeg, '>') != NIL(char))
	       && (brackbeg > message)) {
	    while (*--brackbeg == ' ')
	       ;

	    /* Remove surrounding quotes ? */
	    if ((*brackbeg == '"') && (*author ==  '"')) {
	       *brackbeg = '\0';
	       author++;

	       /* Rather strip trailing spaces here */

	    }
            else {
	       *++brackbeg = '\0';
	    }
	 }
         else {

	    /* 
	     * Check for case 3, <user@domain> -> usr@domain
	     *
	     * Don't need to do this again:
	     * brackbeg = strchr(message, '<');
             */

	    brackend = strchr(message, '>');
	    if ((author == brackbeg) && (brackend != NIL(char))) {
	       author++;
	       *brackend = '\0';
	    }
            else {
	       if ((end = strchr(author, ' ')) != NIL(char)) {
	          *end = '\0';
	       }
	    }
	 }
      }
      /*
       * do a final trimming - just in case the authors name ends
       * in spaces or tabs - it does happen
       */
      end = author + strlen(author) - 1;
      while ((end > author) && ((*end == ' ') || (*end == '\t'))) {
         *end = '\0';
	 end--;
      }
      if (row == -1) {
         sprintf(subjects[count], "%-48.48s %-23.23s", subjects[count], author);
         subjects[count][strlen(subjects[count]) - 1] = '\0';
         ++count;
      }
      else 
         /* Hey - fix this: */
         sprintf((char *) subjects, "%-48.48s %-23.23s", (char *) subjects, author);
   }
   #ifdef DEBUGNNTP
   printf("nntp_connect: end\n");
   #endif


   return;
}

